/* gr_ryn.c
 * RDD, Wed Jan 23 08:28:44 CST 2002
 *
 * Implementations of grammar specific routines 
 * for RYN, RY2, and RY3
 *
 * Mon Jan 21 17:49:12 CST 2002
 * Revised Unambiguous Yarn (II) :	
 *      S -> aS | T | end			(RYN)
 *      T -> Ta | aPb | T aPa' 
 *      P -> aPa' | aS | Ta | T aPa'
 *
 * Variants of above: Wed Apr 04 23:15:35 CST 2002
 * RY2 is Revised Unambiguous Yarn (II) above except:
 *      P -> aPa' | N
 *	N -> aS | Ta | T aPa'
 *
 * RY3 is Revised Unambiguous Yarn (II) except:
 *      P -> aPa' | aNa'
 *	N -> aS | Ta | T aPa'
 *
 * These implementations assume a digitized sequence.
 * See alphabet.c and alphabet.h for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "squid.h"
#include "sqfuncs.h"

#include "cfg.h"
#include "trace.h"
#include "options.h"
#include "alphabet.h"

/* Function: cykInitR
 * Date:     RDD, Wed Jan 23 08:28:44 CST 2002 [St Louis]
 *
 * Purpose:  Initialize CYK fill matrix for RUN and RYN grammars 
 *
 * Args:     
 *	mx	fill matrix (integer log odds form)
 * 	len	sequence length
 *      grammar which grammar to utilize
 *
 * Returns:  -- void -- 
 */
void
cykInitR(int ***mx, INTMOD *model, char *rna, int len, int grammar)
{
   int j, d;

   for (d = 0; d <= len; d++) {
      for (j = d; j < len; j++) {
	 mx[dpS][j][d] = model->transitions[TSE];        /* S -> end */
	 mx[dpT][j][d] = -BIGINT;
	 mx[dpP][j][d] = -BIGINT;
	 if (grammar != RYN) mx[dpN][j][d] = -BIGINT;
      } /* end for j */
      if (d > 0) {
	 mx[dpS][d-1][d] = model->transitions[TSE];
	 mx[dpT][d-1][d] = -BIGINT;
	 mx[dpP][d-1][d] = -BIGINT;
	 if (grammar != RYN) mx[dpN][d-1][d] = -BIGINT;
      }
   } /* end for i */
}

/* Function: cykFillR
 * Date:    Tue Jan 22 15:01:41 CST 2002 [St Louis]
 *
 * Purpose:  Fills CYK matrix for RYN, RY2, and RY3
 *
 * Assumption: Fill matrix already allocated and initialized
 *
 * Args:    
 *	mx	matrix to calculate cyk (as integer log Probs)
 *	sc	parameters 
 *		(integer log form or as scores)
 *	rna	sequence (digitized)
 * 	len	sequence length
 *      grammar which grammar to utilize
 *
 * Notes: Fills matrix correctly for RUN and RYN.
 *      Must fill T before S as one rule in S allows S -> T
 *      and therefore the current cell of T will need to
 *      already known.
 *
 * Returns:  void 
 */
void
cykFillR(int ***mx, INTMOD *pr, char *rna, int len, int grammar)
{
  int d, n, i, j, k;       /* Loop indicies */
  int max;              /* Maximum value seen so far */
  int cursc, cursc2;    /* Current score */
  int stackvalue;	

  /* Recursion */
  for (d = 1; d <= len; d++) {
     for (j = d-1; j < len; j++) {
	i = j - d + 1; max = -BIGINT;

	/* T -> Ta */
	if (j > 0) {
	   cursc = mx[dpT][j-1][d-1] + pr->singles[rna[j]] + pr->transitions[TTT];
	   if (cursc > max) max = cursc;

	   /* T -> aPa' */
	   if (d > 1) {
	      cursc = mx[dpP][j-1][d-2] + pr->pairs[rna[i]][rna[j]] + pr->transitions[TTP];
	      if ((cursc > max) && (lplen(i,j) >= HLEN)) max = cursc; 
	   }
	}

	/* T -> T aPa' */
	for (k = i+1; k < j-1; k++) {
	   cursc = mx[dpT][k][dist(i,k)] + mx[dpP][j-1][dist(k+2,j-1)] 
	      + pr->pairs[rna[k+1]][rna[j]] + pr->transitions[TTB];
	   if ((cursc > max) && (lplen(k+1,j) >= HLEN)) max = cursc; 
	} /* End for loop T -> T aSa' */

	/* T -> Ta | aPa' | T aPa' */
	mx[dpT][j][d] = max;

	max = -BIGINT;
	/* S -> aS */
	cursc = mx[dpS][j][d-1] + pr->singles[rna[i]] + pr->transitions[TSS];
	if (cursc > max) max = cursc; 

	/* S -> T */
	cursc = mx[dpT][j][d] + pr->transitions[TST];
	if (cursc > max) max = cursc;

	/* S -> aS | T | end */
	mx[dpS][j][d] = max;

	if (grammar != RYN) {
	   /* N -> aS | Tb | TaPa' */
	   max = -BIGINT;

	   /* N -> aS */
	   cursc = mx[dpS][j][d-1] + pr->singles[rna[i]] + pr->transitions[TNL];
	   if (cursc > max) max = cursc; 

	   /* N -> Tb */
	   if (j > 0) {
	      cursc = mx[dpT][j-1][d-1] + pr->singles[rna[j]] + pr->transitions[TNR];
	      if (cursc > max) max = cursc;
	   }

	   /* N -> T aPa' */ 
	   for (k = i+1; k < j-1; k++) {
	      cursc = mx[dpT][k][dist(i,k)] + mx[dpP][j-1][dist(k+2,j-1)] 
		 + pr->pairs[rna[k+1]][rna[j]] + pr->transitions[TNB];
	      if ((cursc > max) && (lplen(k+1,j) >= HLEN)) max = cursc; 
	   }
	   mx[dpN][j][d] = max;
	}

	max = -BIGINT;
	/* Stacking! Can't be in P state at edges!  */
	if ((i == 0) || (j == len-1)) {
	   stackvalue = -BIGINT;
	} else {
	   stackvalue = pr->stack[idx(rna[i-1],rna[j+1])][rna[i]][rna[j]];
	}
	/* P -> aPa */
	if ((j > 0) && ( d > 1) && (lplen(i,j) >= HLEN)) {
	   cursc = mx[dpP][j-1][d-2] + stackvalue + pr->transitions[TPP];
	}
	if (cursc > max) max = cursc;

	if (grammar == RYN) {
	   /* P -> aS */
	   cursc = mx[dpS][j][d-1] + pr->singles[rna[i]] + pr->transitions[TPS];
	   if (cursc > max) max = cursc; 

	   /* P -> Tb */
	   if (j > 0) {
	      cursc = mx[dpT][j-1][d-1] + pr->singles[rna[j]] + pr->transitions[TPT];
	      if (cursc > max) max = cursc;
	   }

	   /* P -> T aPa' */
	   for (k = i+1; k < j-1; k++) {
	      cursc = mx[dpT][k][dist(i,k)] + mx[dpP][j-1][dist(k+2,j-1)] 
		 + pr->pairs[rna[k+1]][rna[j]] + pr->transitions[TPB];
	      if ((cursc > max) && (lplen(k+1,j) >= HLEN)) max = cursc; 
	   }

	   /* P -> aPb | aS | Tb | T aPa' */
	   mx[dpP][j][d] = max;

	} else {
	   if (grammar == RY2) {
	      /* P -> N */
	      cursc2 = mx[dpN][j][d] + pr->transitions[TPN];
	   } else {
	      /* P -> aNa' */
	      if ((j > 0) && (d > 1) && (lplen(i,j) >= HLEN)) {
		 cursc2 = mx[dpN][j-1][d-2] + stackvalue + pr->transitions[TPN];
	      } else {
		 cursc2 = -BIGINT;
	      }
	   }
	   if (cursc > cursc2) {
	      mx[dpP][j][d] = cursc;
	   } else {
	      mx[dpP][j][d] = cursc2;
	   }
	}
     }
  }
}

/* Function: cykTraceR
 * Date:     RDD, Tue Jan 22 15:24:12 CST 2002 [St Louis]
 *
 * Purpose:  Build traceback tree for RUN and RYN 
 *
 * Args:
 *	mx	matrix in which to calculate cyk
 *	pr	parameters in integer form
 *	rna	sequence 
 * 	len	sequence length
 *
 * Returns: traceback tree
 */
struct trace_s *
cykTraceR(int ***mx, INTMOD *pr, char *rna, int len, int grammar)
{
  int d, i, j, k, mtx;   /* Indices */
  struct trace_s *parsetree;     /* Head of parse tree */
  struct trace_s *curr;          /* Current node in parse tree */
  struct tracestack_s *stack;    /* Stack for doing traceback */

  /* Initialization of Traceback stuff */
  parsetree = InitTrace();
  stack = InitTracestack();

  /* Push initialization elements onto stack */
  PushTracestack(stack, AttachTrace (parsetree, dpS, 0, (len -1), TRE));

  /* Repeat until stack is empty */
  while (curr = PopTracestack(stack)) {
    /* Set i and j from items in stack */
    i = curr->emitl; j = curr->emitr; mtx = curr->nonterminal;
    d = j - i + 1;

    if (i > j) { 		/* S(R) -> end */
      curr->emitl = -1; curr->emitr = -1; 
      curr->transition = TSE; curr->nonterminal = dpS; 
      continue;
    } else if (mtx == dpS) {
      /* S -> aS | T | end */
      if ((mx[dpS][j][d-1] + pr->singles[rna[i]] + pr->transitions[TSS]) == mx[dpS][j][d]) {
        curr->emitr = -1; curr->transition = TSS; 
        PushTracestack(stack, AttachTrace (curr, dpS, i+1, j, TRE));
      } else {
	/* Assumes: ((mx[dpT][j][d] + pr->transitions[TST]) == mx[dpS][j][d]) */
        curr->emitr = -1; curr->emitl = -1; curr->transition = TST; 
        PushTracestack(stack, AttachTrace (curr, dpT, i, j, TRE));
      }

    } else if (mtx == dpT) {
      /* T -> Tb | aPa' | T aPa' */
      if ((mx[dpT][j-1][d-1] + pr->singles[rna[j]] + pr->transitions[TTT]) == mx[dpT][j][d]) {
        curr->emitl = -1; curr->transition = TTT; 
        PushTracestack(stack, AttachTrace (curr, dpT, i, j-1, TRE));
      } else if ((lplen(i,j) >= HLEN) && ((mx[dpP][j-1][d-2] 
		  + pr->pairs[rna[i]][rna[j]] + pr->transitions[TTP]) 
	       == mx[dpT][j][d])) {
        curr->transition = TTP;
        PushTracestack(stack, AttachTrace (curr, dpP, i+1, j-1, TRE));
      } else {
        for (k = i+1; k < j-1; k++) {
	   if (lplen(k+1,j) >= HLEN) {
	      if ((mx[dpT][k][dist(i,k)] + mx[dpP][j-1][dist(k+2,j-1)] 
		       + pr->pairs[rna[k+1]][rna[j]] + 
		       pr->transitions[TTB]) == mx[dpT][j][d]) {
		 curr->emitl = k+1; curr->transition = TTB;
		 PushTracestack(stack, AttachTrace (curr, dpP, k+2, j-1, TRE));
		 PushTracestack(stack, AttachTrace (curr, dpT, i, k, TRE));
		 k = j;	/* Short circuit */
	      } /* End if */
	   }
        } /* End for (bifurcation points) */
      } 

    } else if (mtx == dpP) {
      if (grammar == RYN) {
	/* P -> aPa' | aS | Ta | T aPa' */
	if ((lplen(i,j) >= HLEN) && ((mx[dpP][j-1][d-2] 
		    + pr->stack[idx(rna[i-1],rna[j+1])][rna[i]][rna[j]] 
		    + pr->transitions[TPP]) == mx[dpP][j][d])) {
	  curr->transition = TPP;
	  PushTracestack(stack, AttachTrace (curr, dpP, i+1, j-1, TRE));
	} else  if ((mx[dpS][j][d-1] + pr->singles[rna[i]] + pr->transitions[TPS]) 
	      				== mx[dpP][j][d]) {
	  curr->emitr = -1; curr->transition = TPS; 
	  PushTracestack(stack, AttachTrace (curr, dpS, i+1, j, TRE));
	} else if ((mx[dpT][j-1][d-1] + pr->singles[rna[j]] + pr->transitions[TPT]) 
	      				== mx[dpP][j][d]) {
	  curr->emitl = -1; curr->transition = TPT; 
	  PushTracestack(stack, AttachTrace (curr, dpT, i, j-1, TRE));
	} else {
	  for (k = i+1; k < j-1; k++) {
	     if (lplen(k+1,j) >= HLEN) {
		if ((mx[dpT][k][dist(i,k)] + mx[dpP][j-1][dist(k+2,j-1)] 
			 + pr->pairs[rna[k+1]][rna[j]] 
			 + pr->transitions[TPB]) == mx[dpP][j][d]) {
		   curr->emitl = k+1; curr->transition = TPB;
		   PushTracestack(stack, AttachTrace (curr, dpP, k+2, j-1, TRE));
		   PushTracestack(stack, AttachTrace (curr, dpT, i, k, TRE));
		   k = j;	/* Short circuit */
		} /* End if */
	     }
	  } /* End for (bifurcation points) */
	} 
      } else {
	if ((lplen(i,j) >= HLEN) && ((mx[dpP][j-1][d-2] 
		    + pr->stack[idx(rna[i-1],rna[j+1])][rna[i]][rna[j]] 
		    + pr->transitions[TPP]) == mx[dpP][j][d])) {
	  curr->transition = TPP;
	  PushTracestack(stack, AttachTrace (curr, dpP, i+1, j-1, TRE));
        } else {
          curr->transition = TPN;
          if (grammar == RY3) {
            PushTracestack(stack, AttachTrace (curr, dpN, i+1, j-1, TRE)); 
          } else { 
	    curr->emitl = -1; curr->emitr = -1; 
            PushTracestack(stack, AttachTrace (curr, dpN, i, j, TRE)); 
          }
        }
      }
    } else if (mtx == dpN) {
      if ((mx[dpS][j][d-1] + pr->singles[rna[i]] + pr->transitions[TNL]) == mx[dpN][j][d]) {
        curr->emitr = -1; curr->transition = TNL; 
        PushTracestack(stack, AttachTrace (curr, dpS, i+1, j, TRE));
      } else if ((mx[dpT][j-1][d-1] + pr->singles[rna[j]] + pr->transitions[TNR]) 
	    				== mx[dpN][j][d]) {
        curr->emitl = -1; curr->transition = TNR; 
        PushTracestack(stack, AttachTrace (curr, dpT, i, j-1, TRE));
      } else {
	 for (k = i+1; k < j-1; k++) {
	    if (lplen(k+1, j) >= HLEN) {
	       if ((mx[dpT][k][dist(i,k)] + mx[dpP][j-1][dist(k+2,j-1)] 
			+ pr->pairs[rna[k+1]][rna[j]] 
			+ pr->transitions[TNB]) == mx[dpN][j][d]) {
		  curr->emitl = k+1; curr->transition = TNB;
		  PushTracestack(stack, AttachTrace (curr, dpP, k+2, j-1, TRE));
		  PushTracestack(stack, AttachTrace (curr, dpT, i, k, TRE));
		  k = j;	/* Short circuit */
	       } /* End if */
	    }
        } /* End for (bifurcation points) */
      } 
    } else {
      printf("ERROR!! Nonterminal %d unkown in RUN/RYN traceback!\n", mtx);
    }
  }
  FreeTracestack(stack);
  return parsetree;
}

/* Function: khs2traceRYN
 * Date:     RDD, Tue Apr 23 09:39:41 2002 [St. Louis]
 *
 * Purpose:  Convert KHS format into properly 
 * 	labeled traceback tree for RYN grammar
 *
 * Args:     
 *	dolist  Traceback tree to build from KHS	
 *	ct	KHS info represented as ct format
 */
khs2traceRYN(struct tracestack_s *dolist, int *ct)
{
  struct trace_s      *cur;
  struct trace_s      *tr;
  struct trace_s      *prv, *nxt;
  char left; char right;
  int  i,j, mid;

  while ((cur = PopTracestack(dolist)) != NULL) {
    i = cur->emitl; j = cur->emitr;

    if (i > j) {
      cur->transition = TSE; cur->emitl = -1; cur->emitr = -1; 
    } else if (ct[i] == -1) {
      if (cur->nonterminal == dpS) {
        cur->transition = TSS; 
      } else {
        cur->transition = TPS; 
      }
      cur->emitr = -1;
      PushTracestack(dolist, AttachTrace(cur, dpS, i+1, j, TRE));
    } else if (ct[j] == -1) {
      if (cur->nonterminal == dpS) {
        cur->transition = TST; cur->emitl = -1; cur->emitr = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j, TRE));
      } else {
        if (cur->nonterminal == dpT) {
          cur->transition = TTT; 
        } else {
          cur->transition = TPT; 
        }
        cur->emitl = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j-1, TRE));
      }
    } else if (ct[i] == j) {
      if (i+1 == j) { return 0;	/* Minimum hp length 1 nt */ }
      if (cur->nonterminal == dpS) {
        cur->transition = TST; cur->emitl = -1; cur->emitr = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j, TRE));
      } else {
        if (cur->nonterminal == dpT) {
          cur->transition = TTP;
        } else {
          cur->transition = TPP;
        }
        PushTracestack(dolist, AttachTrace(cur, dpP, i+1, j-1, TRE));
      }
    } else {
      if (cur->nonterminal == dpS) {
        cur->transition = TST; cur->emitl = -1; cur->emitr = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j, TRE));
      } else {
	if (ct[j]+1 == j) { return 0; /* Min hp length 1 nt */ }
        if (cur->nonterminal == dpT) {
          cur->transition = TTB; 
        } else {
          cur->transition = TPB; 
        }
        cur->emitl = ct[j];
        PushTracestack(dolist, AttachTrace(cur, dpP, ct[j]+1, j-1, TRE));
        PushTracestack(dolist, AttachTrace(cur, dpT, i, ct[j]-1, TRE));
      }
    }
  }
  return 1;
}

/* Function: khs2traceRY2
 * Date:     RDD, Tue Apr 23 09:39:55 2002 [St. Louis]
 *
 * Purpose:  Convert KHS format into properly 
 * 	labeled traceback tree for RY2 grammars
 *
 * Args:     
 *	dolist  Traceback tree to build from KHS	
 *	ct	KHS info represented as ct format
 */
int
khs2traceRY2(struct tracestack_s *dolist, int *ct)
{
  struct trace_s      *cur;
  struct trace_s      *tr;
  struct trace_s      *prv, *nxt;
  char left; char right;
  int  i,j, mid;

  while ((cur = PopTracestack(dolist)) != NULL) {
    i = cur->emitl; j = cur->emitr;

    if (i > j) {
      cur->transition = TSE; cur->emitl = -1; cur->emitr = -1; 
    } else if (ct[i] == -1) {
       if (cur->nonterminal == dpP) {
	  cur->transition = TPN; 
	  cur->emitr = -1; cur->emitl = -1;
	  PushTracestack(dolist, AttachTrace(cur, dpN, i, j, TRE));
       } else {
	  if (cur->nonterminal == dpS) {
	     cur->transition = TSS; 
	  } else {
	     cur->transition = TNL; 
	  }
	  cur->emitr = -1;
	  PushTracestack(dolist, AttachTrace(cur, dpS, i+1, j, TRE));
       }
    } else if (ct[j] == -1) {
      if (cur->nonterminal == dpS) {
        cur->transition = TST; cur->emitl = -1; cur->emitr = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j, TRE));
      } else if (cur->nonterminal == dpP) {
        cur->transition = TPN; cur->emitl = -1; cur->emitr = -1;
        PushTracestack(dolist, AttachTrace(cur, dpN, i, j, TRE));
      } else {
        if (cur->nonterminal == dpT) {
          cur->transition = TTT; 
        } else {
          cur->transition = TNR; 
        }
        cur->emitl = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j-1, TRE));
      }
    } else if (ct[i] == j) {
      if (cur->nonterminal == dpS) {
        cur->transition = TST; cur->emitl = -1; cur->emitr = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j, TRE));
      } else {
	 if (i+1 == j) { return 0;	/* Minimum hp length 1 nt */ }
	 if (cur->nonterminal == dpT) {
	    cur->transition = TTP;
	 } else {
	    cur->transition = TPP;
	 }
	 PushTracestack(dolist, AttachTrace(cur, dpP, i+1, j-1, TRE));
      }
    } else {
      if (cur->nonterminal == dpS) {
        cur->transition = TST; cur->emitl = -1; cur->emitr = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j, TRE));
      } else if (cur->nonterminal == dpP) {
        cur->transition = TPN; cur->emitl = -1; cur->emitr = -1;
        PushTracestack(dolist, AttachTrace(cur, dpN, i, j, TRE));
      } else {
	if (ct[j]+1 == j) { return 0; /* Min hp length 1 nt */ }
        if (cur->nonterminal == dpT) {
          cur->transition = TTB; 
        } else {
          cur->transition = TNB; 
        }
        cur->emitl = ct[j];
        PushTracestack(dolist, AttachTrace(cur, dpP, ct[j]+1, j-1, TRE));
        PushTracestack(dolist, AttachTrace(cur, dpT, i, ct[j]-1, TRE));
      }
    }
  }
  return 1;
}

/* Function: khs2traceRY3
 * Date:     RDD, Sat Dec 13 12:05:22 CST 2003 [St. Louis]
 *
 * Purpose:  Convert KHS format into properly 
 * 	labeled traceback tree for RY3 grammars
 *
 * Args:     
 *	dolist  Traceback tree to build from KHS	
 *	ct	KHS info represented as ct format
 *	tolerate  Do we manage problems with the KHS or
 *		do we throw an error?
 */
int
khs2traceRY3(struct tracestack_s *dolist, int *ct, int tolerate)
{
  struct trace_s      *cur;
  struct trace_s      *tr;
  struct trace_s      *prv, *nxt;
  char left; char right;
  int  i,j, mid;

  while ((cur = PopTracestack(dolist)) != NULL) {
    i = cur->emitl; j = cur->emitr;

    if (i > j) {
      cur->transition = TSE; cur->emitl = -1; cur->emitr = -1; 
    } else if (ct[i] == -1) {
       if (cur->nonterminal == dpP) {
	  if (tolerate) {
	     cur->transition = TPN; 
	     cur->emitr = -1; cur->emitl = -1;
	     PushTracestack(dolist, AttachTrace(cur, dpN, i, j, TRE));
	  } else {
	     return 0;	/* Lone pairs not permissible */
	  }
       } else {
	  if (cur->nonterminal == dpS) {
	     cur->transition = TSS; 
	  } else {
	     cur->transition = TNL; 
	  }
	  cur->emitr = -1;
	  PushTracestack(dolist, AttachTrace(cur, dpS, i+1, j, TRE));
       }
    } else if (ct[j] == -1) {
      if (cur->nonterminal == dpS) {
        cur->transition = TST; cur->emitl = -1; cur->emitr = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j, TRE));
      } else if (cur->nonterminal == dpP) {
	 if (tolerate) {
	    cur->transition = TPN; cur->emitl = -1; cur->emitr = -1;
	    PushTracestack(dolist, AttachTrace(cur, dpN, i, j, TRE));
	 } else {
	    return 0;	/* Lone pairs not permissible */
	 }
      } else {
        if (cur->nonterminal == dpT) {
          cur->transition = TTT; 
        } else {
          cur->transition = TNR; 
        }
        cur->emitl = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j-1, TRE));
      }
    } else if (ct[i] == j) {
      if (cur->nonterminal == dpS) {
        cur->transition = TST; cur->emitl = -1; cur->emitr = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j, TRE));
      } else {
	 if (i+1 == j) { return 0;	/* Minimum hp length 1 nt */ }
	 if ((cur->nonterminal == dpP) && (ct[i+1] != j-1)) {
	    cur->transition = TPN;
	    PushTracestack(dolist, AttachTrace(cur, dpN, i+1, j-1, TRE));
	 } else {
	    if (cur->nonterminal == dpT) {
	       cur->transition = TTP;
	    } else {
	       cur->transition = TPP;
	    }
	    PushTracestack(dolist, AttachTrace(cur, dpP, i+1, j-1, TRE));
	 }
      }
    } else {
      if (cur->nonterminal == dpS) {
        cur->transition = TST; cur->emitl = -1; cur->emitr = -1; 
        PushTracestack(dolist, AttachTrace(cur, dpT, i, j, TRE));
      } else if (cur->nonterminal == dpP) {
	 if (tolerate) {
	    cur->transition = TPN; cur->emitl = -1; cur->emitr = -1;
	    PushTracestack(dolist, AttachTrace(cur, dpN, i, j, TRE));
	 } else {
	    return 0;	/* Lone pairs not permissible */
	 }
      } else {
	if (ct[j]+1 == j) { return 0; /* Min hp length 1 nt */ }
        if (cur->nonterminal == dpT) {
          cur->transition = TTB; 
        } else {
          cur->transition = TNB; 
        }
        cur->emitl = ct[j];
        PushTracestack(dolist, AttachTrace(cur, dpP, ct[j]+1, j-1, TRE));
        PushTracestack(dolist, AttachTrace(cur, dpT, i, ct[j]-1, TRE));
      }
    }
  }
  return 1;
}

/* Function: analyzeTraceR2 
 * NOTE: RYN uses RUN's analyzeTraceR, this is RY2 and RY3 only.
 * Date:     RDD, Tue May 14 16:01:32 CDT 2002 [St Louis]
 *
 * Purpose: Parse a traceback tree for information.
 *      Either for counts (training).
 *      Or for determining tracebcak score.
 *
 * Args:
 *      seq   - sequence corresponding to the trace
 *      len   - length of the sequence
 *      tr    - traceback to count
 *      cfg   - model, counts form, to add counts to
 *      count - if TRUE this is TraceCount;
 *              if FALSE this is TraceScore
 *
 * Returns:  score of current traceback (in TraceScore mode)
 *      Should return zero in TraceCount mode
 */
int
analyzeTraceR2(char *seq, int len, struct trace_s *tr, INTMOD *cfg, int grammar, int count)
{
  struct tracestack_s *dolist;  /* stack for traversal of traceback tree */
  struct trace_s      *cur;     /* current node in the tree              */
  char *iseq;                   /* sequence with 0,1,2,3,4 in place of A,C,G,U,N */
  int   cst, nst, j;
  char left; char right;
  int score;

 /* Digitize Sequence setup */
  SetAlphabet(hmmNUCLEIC);

  iseq = DigitizeSequence(seq, len);

  dolist = InitTracestack();
  PushTracestack(dolist, tr->nxtl);
  score = 0;

  while ((cur = PopTracestack(dolist)) != NULL) {
    if (cur->transition < NTRANS) {
      if (count) cfg->transitions[cur->transition] += 1;
      else score += cfg->transitions[cur->transition];
      if (cur->transition == TSE) {	/* do nothing on ENDS */
	continue;
      } else if (cur->transition == TTB) {
        if (count) cfg->pairs[iseq[cur->emitl]][iseq[cur->emitr]] += 1;
        else score += cfg->pairs[iseq[cur->emitl]][iseq[cur->emitr]];
	PushTracestack(dolist, cur->nxtr);
	PushTracestack(dolist, cur->nxtl);
      } else if (cur->transition == TNB) {
        if (grammar == RY3) {
          if (count) cfg->pairs[iseq[cur->emitl]][iseq[cur->emitr]] += 1;
          else score += cfg->pairs[iseq[cur->emitl]][iseq[cur->emitr]];
        }
	PushTracestack(dolist, cur->nxtr);
	PushTracestack(dolist, cur->nxtl);
      } else if ((cur->emitl == -1) && (cur->emitr == -1)) { /* No emission */
	PushTracestack(dolist, cur->nxtl);
      } else {
	 if (cur->emitl == -1) {
	    if (iseq[cur->emitr] > ALPHA) {
	       /* Do nothing, contains an ambiguous base! */
	       /* Therfore must back out of counting transition?  */
	       if (count) cfg->transitions[cur->transition] -= 1;
	    } else {
	       if (count) cfg->singles[iseq[cur->emitr]] += 1;
	       else score += cfg->singles[iseq[cur->emitr]];
	    }

	    /* Emit to left */
	 } else if (cur->emitr == -1) {
	    if (iseq[cur->emitl] > ALPHA) {
	       /* Do nothing, contains an ambiguous base! */
	       /* Therfore must back out of counting transition?  */
	       if (count) cfg->transitions[cur->transition] -= 1;
	    } else {
	       if (count) cfg->singles[iseq[cur->emitl]] += 1;
	       else score += cfg->singles[iseq[cur->emitl]];
	    }

	/* Emit aligned pair -- stacking model */
        } else if (cur->nonterminal == dpP) {
          if ((iseq[cur->emitl -1] > ALPHA) || (iseq[cur->emitr +1] > ALPHA)) {
		/* Do nothing, contains an ambiguous base! 
		 * Therfore must back out of counting transition!
 		 */
	    if (count) cfg->transitions[cur->transition] -= 1;
          } else {
	    if (count) {
              cfg->stack[idx(iseq[cur->emitl - 1],iseq[cur->emitr + 1])]
			[iseq[cur->emitl]][iseq[cur->emitr]] += 1;
	      /* By also adding to pairs, it is like you are averaging the 
	       * pair emissions over all pairs (including stacks) */
	      cfg->pairs[iseq[cur->emitl]][iseq[cur->emitr]] += 1;
            } else {
              score += cfg->stack[idx(iseq[cur->emitl - 1],iseq[cur->emitr + 1])]
			[iseq[cur->emitl]][iseq[cur->emitr]];
            }
          }

	/* Emit aligned pair -- no stacking */
	} else {
	  if (count) cfg->pairs[iseq[cur->emitl]][iseq[cur->emitr]] += 1;
	  else score += cfg->pairs[iseq[cur->emitl]][iseq[cur->emitr]];
	}
	PushTracestack(dolist, cur->nxtl);
      }
    } else {
      PushTracestack(dolist, cur->nxtl);
    }
  }
  free(iseq);
  FreeTracestack(dolist);
  return score;
}
/* Function: dTraceScoreR2 
 * NOTE: RYN uses RUN's dTraceScoreR, this is RY2 and RY3 only.
 * Date:     Mon Oct 27 17:00:37 CST 2003 [St Louis]
 *
 * Purpose: Parse a traceback tree for determining tracebcak score.
 * 	Needed for ambiguity testing.
 *
 * Args:
 *      seq   - sequence corresponding to the trace
 *      len   - length of the sequence
 *      tr    - traceback to count
 *      cfg   - model in float log form
 *      ret_val - score of traceback
 *
 * Returns:  void
 */
void
dTraceScoreR2(char *seq, int len, struct trace_s *tr, PROBMOD *cfg, 
      int grammar, double *ret_val)
{
  struct tracestack_s *dolist;  /* stack for traversal of traceback tree */
  struct trace_s      *cur;     /* current node in the tree              */
  char *iseq;                   /* sequence with 0,1,2,3,4 in place of A,C,G,U,N */
  int   cst, nst, j;
  char left; char right;
  double score;

 /* Digitize Sequence setup */
  SetAlphabet(hmmNUCLEIC);

  iseq = DigitizeSequence(seq, len);

  dolist = InitTracestack();
  PushTracestack(dolist, tr->nxtl);
  score = 0.;

  while ((cur = PopTracestack(dolist)) != NULL) {
    if (cur->transition < NTRANS) {
      score += cfg->transitions[cur->transition];
      if (cur->transition == TSE) {	/* do nothing on ENDS */
	continue;
      } else if (cur->transition == TTB) {
        score += cfg->pairs[iseq[cur->emitl]][iseq[cur->emitr]];
	PushTracestack(dolist, cur->nxtr);
	PushTracestack(dolist, cur->nxtl);
      } else if (cur->transition == TNB) {
        if (grammar != RYN) {
          score += cfg->pairs[iseq[cur->emitl]][iseq[cur->emitr]];
        }
	PushTracestack(dolist, cur->nxtr);
	PushTracestack(dolist, cur->nxtl);
      } else if ((cur->emitl == -1) && (cur->emitr == -1)) { /* No emission */
	PushTracestack(dolist, cur->nxtl);
      } else {
	 if (cur->emitl == -1) {
	    if (iseq[cur->emitr] > ALPHA) {
	       /* Do nothing, contains an ambiguous base! */
	    } else {
	       score += cfg->singles[iseq[cur->emitr]];
	    }

	    /* Emit to left */
	 } else if (cur->emitr == -1) {
	    if (iseq[cur->emitl] > ALPHA) {
	       /* Do nothing, contains an ambiguous base! */
	    } else {
	       score += cfg->singles[iseq[cur->emitl]];
	    }

	    /* Emit aligned pair -- stacking model */
	 } else if (cur->nonterminal == dpP) {
          if ((iseq[cur->emitl -1] > ALPHA) || (iseq[cur->emitr +1] > ALPHA)) {
		/* Do nothing, contains an ambiguous base! 
		 * Therfore must back out of counting transition!
 		 */
          } else {
              score += cfg->stack[idx(iseq[cur->emitl - 1],iseq[cur->emitr + 1])]
			[iseq[cur->emitl]][iseq[cur->emitr]];
          }

	/* Emit aligned pair -- no stacking */
	} else {
	  score += cfg->pairs[iseq[cur->emitl]][iseq[cur->emitr]];
	}
	PushTracestack(dolist, cur->nxtl);
      }
    } else {
      PushTracestack(dolist, cur->nxtl);
    }
  }
  free(iseq);
  FreeTracestack(dolist);
  *ret_val = score;
}

/************************ Inside *****************************/

/* Function: cinsInitR
 * Date:     RDD, Sun Oct  5 14:50:00 CDT 2003 [St Louis]
 *
 * Purpose:  Initialize cInside fill matrix 
 *
 * Args:     
 *	mx	fill matrix (log odds form)
 * 	len	sequence length
 *      grammar which grammar to utilize
 *
 * Returns:  -- void -- 
 */
void
cinsInitR(double ***mx, PROBMOD *model, char *rna, int len, int grammar)
{
   int j, d;

   for (d = 0; d <= len; d++) {
      for (j = d; j < len; j++) {
	 mx[dpS][j][d] = model->transitions[TSE];        /* S -> end */
	 mx[dpT][j][d] = -BIGFLOAT;
	 mx[dpP][j][d] = -BIGFLOAT;
	 if (grammar != RYN) mx[dpN][j][d] = -BIGFLOAT;
      } /* end for j */
      if (d > 0) {
	 mx[dpS][d-1][d] = model->transitions[TSE];
	 mx[dpT][d-1][d] = -BIGFLOAT;
	 mx[dpP][d-1][d] = -BIGFLOAT;
	 if (grammar != RYN) mx[dpN][d-1][d] = -BIGFLOAT;
      }
   } /* end for i */
}

/* Function: cinsFillR
 * Date:    Sun Oct  5 14:51:27 CDT 2003 [St Louis]
 *
 * Purpose:  Fills cInside matrix for RYN, RY2, and RY3
 *
 * Assumption: Fill matrix already allocated and initialized
 *
 * Args:    
 *	mx	matrix to calculate cyk (log Probs)
 *	sc	parameters (log form )
 *	rna	sequence (digitized)
 * 	len	sequence length
 *      grammar which grammar to utilize
 *
 * Notes: Fills matrix correctly for RUN and RYN.
 *      Must fill T before S as one rule in S allows S -> T
 *      and therefore the current cell of T will need to
 *      already known.
 *
 * Returns:  void 
 */
void
cinsFillR(double ***mx, PROBMOD *pr, char *rna, int len, int grammar, int *ss)
{
  int d, n, i, j, k;       /* Loop indicies */
  double cursc;    /* Current score */
  double stackvalue;	

  /* Recursion */
  for (d = 1; d <= len; d++) {
     for (j = d-1; j < len; j++) {
	i = j - d + 1; 

	cursc = -BIGFLOAT;
	/* T -> Ta | aPa' | T aPa' */
	if (j > 0) {
	   if (ss[j] == -1) cursc = mx[dpT][j-1][d-1] + pr->singles[rna[j]] + pr->transitions[TTT];

	   /* T -> aPa' */
	   if (ss[i] == j) {
	      if (d > 1) {
		 cursc = DLogsum(mx[dpP][j-1][d-2] + pr->pairs[rna[i]][rna[j]] + pr->transitions[TTP], cursc);
	      }
	   }
	}

	/* T -> T aPa' */
	for (k = i+1; k < j-1; k++) {
	   if (ss[k+1] == j) {
	      cursc = DLogsum(mx[dpT][k][dist(i,k)] + mx[dpP][j-1][dist(k+2,j-1)] 
		 + pr->pairs[rna[k+1]][rna[j]] + pr->transitions[TTB], cursc);
	   }
	} 

	/* T -> Ta | aPa' | T aPa' */
	mx[dpT][j][d] = cursc;

	cursc = -BIGFLOAT;
	/* S -> aS | T | end */
	if (ss[i] == -1) cursc = mx[dpS][j][d-1] + pr->singles[rna[i]] + pr->transitions[TSS];
	cursc = DLogsum(mx[dpT][j][d] + pr->transitions[TST], cursc);
	mx[dpS][j][d] = cursc;

	if (grammar != RYN) {
	   /* N -> aS | Ta | TaPa' */
	   cursc = -BIGFLOAT;

	   /* N -> aS */
	   if (ss[i] == -1) cursc = mx[dpS][j][d-1] + pr->singles[rna[i]] + pr->transitions[TNL];

	   /* N -> Ta */
	   if (ss[j] == -1) {
	      if (j > 0) {
		 cursc = DLogsum(mx[dpT][j-1][d-1] + pr->singles[rna[j]] + pr->transitions[TNR], cursc);
	      }
	   }

	   /* N -> T aPa' */ 
	   for (k = i+1; k < j-1; k++) {
	      if (ss[k+1] == j) {
		 cursc = DLogsum(mx[dpT][k][dist(i,k)] + mx[dpP][j-1][dist(k+2,j-1)] 
		    + pr->pairs[rna[k+1]][rna[j]] + pr->transitions[TNB], cursc);
	      }
	   }
	   /* N -> aS | Ta | TaPa' */
	   mx[dpN][j][d] = cursc;
	}

 	/* P -> aPa' | aS | Ta | T aPa'  or   P -> aPa' | N   or   P -> aPa | aNa' */
	cursc = -BIGFLOAT;
	/* Stacking! Can't be in P state at edges!  */
	if ((i == 0) || (j == len-1)) {
	   stackvalue = -BIGFLOAT;
	} else {
	   stackvalue = pr->stack[idx(rna[i-1],rna[j+1])][rna[i]][rna[j]];
	}
	/* P -> aPa */
	if (ss[i] == j) {
	   if ((j > 0) && ( d > 1)) {
	      cursc = mx[dpP][j-1][d-2] + stackvalue + pr->transitions[TPP];
	   }
	}

	if (grammar == RYN) {
	   /* P -> aS */
	   if (ss[i] == -1) {
	      cursc = DLogsum(mx[dpS][j][d-1] + pr->singles[rna[i]] + pr->transitions[TPS], cursc);
	   }

	   /* P -> Ta */
	   if (j > 0) {
	      if (ss[j] == -1) 
		 cursc = DLogsum(mx[dpT][j-1][d-1] + pr->singles[rna[j]] + pr->transitions[TPT], cursc);
	   }

	   /* P -> T aPa' */
	   for (k = i+1; k < j-1; k++) {
	      if (ss[k+1] == j) {
		 cursc = DLogsum(mx[dpT][k][dist(i,k)] + mx[dpP][j-1][dist(k+2,j-1)] 
		    + pr->pairs[rna[k+1]][rna[j]] + pr->transitions[TPB], cursc);
	      }
	   }

	   /* P -> aPb | aS | Tb | T aPa' */
	   mx[dpP][j][d] = cursc;

	} else {
	   if (grammar == RY2) {
	      /* P -> N */
	      cursc = DLogsum(mx[dpN][j][d] + pr->transitions[TPN], cursc);
	   } else {
	      /* P -> aNa' */
	      if (ss[i] == j) {
		 if ((j > 0) && (d > 1)) {
		    cursc = DLogsum(mx[dpN][j-1][d-2] + stackvalue + pr->transitions[TPN], cursc);
		 }
	      } 
	   }
	   /* P -> aPa' | aNa   or   aPa' | N */
	   mx[dpP][j][d] = cursc;
	}
     }
  }
}
