/* khs.c 
 *   SRE, Sun Aug 28 10:39:18 1994
 *   RDD, Tue Oct  2 14:56:30 CDT 2001
 *   RDD, Tue Feb 18 10:33:39 CST 2003
 *
 * Representation of secondary structure and secondary structural 
 * alignments using a variant of Danielle Konings' string notation.
 * See: Konings and Hogeweg, J. Mol. Biol. 207:597-614 1989
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "squid.h"
#include "cfg.h"
#include "consan.h"
#include "options.h"
#include "trace.h"
#include "grammars.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif


/* Function: calcConsStruct 
 * 
 * Purpose:  Given two aligned structure annotation lines, 
 * 	determine the consensus structure between them.
 *
 * Args:     
 * 	ss1	first sequence structure
 * 	seq1	first sequence
 * 	ss2	second sequence structure
 * 	seq2	second sequence
 * 	len	aligned length of two structures/sequences
 * 
 * Return:   1 on success, 0 on failure.
 * 	ret_css	consensus structure (allocated here; free elsewhere)
 */
int
calcConsStruct(char *ss1, char *seq1, char *ss2, char *seq2, 
      int len, char **ret_css)
{
  int *ct1, *ct2;
  int i;
  char *css;

  if ((ss1 == NULL) || (ss2 == NULL)) return FALSE; 

  if (! KHS2ct(ss1, len, FALSE, &ct1))
      { printf("[bad first structure]\n"); return (FALSE);}
  if (! KHS2ct(ss2, len, FALSE, &ct2))
      { printf("[bad second structure]\n"); return (FALSE); }

  css = (char *) malloc(sizeof(char)*len+1);

  for (i = 0; i < len; i++) {
     if (ct1[i] == ct2[i]) { css[i] = ss1[i]; } 
     else { css[i] = '.'; }
  }
  css[len] = '\0';

  if (!(VerifyKHS("built", css, FALSE) == TRUE)) 
      { printf("[bad built structure]\n"); return (FALSE); }

  free(ct1);
  free(ct2);
  *ret_css = css;
  return 1;
}

/* Function: khs2trace
 * 
 * Purpose:  Create a trace structure from a KH structure
 *            string. This is very similar to tracebacks
 *            of a dynamic programming matrix.
 *
 * This function calls the appropriate grammar specific version.
 * 
 * Args:     
 * 
 * Return:   1 on success, 0 on failure.
 *           Caller frees ret_tr with FreeTrace().
 */
int
khs2trace(char *ss, int len, char *seqX, char *seqY, 
      int grammar, struct trace_s **ret_tr)
{
   struct tracestack_s *dolist;
   struct trace_s      *tr;
   int  validstruct;
   int *ct;
   int debug = FALSE;
   int delenX, delenY;

   if (! KHS2ct(ss, len, FALSE, &ct)) {
      Warn("khs2trace Structure string is inconsistent"); 
      return 0;
   }

   validstruct = TRUE;
   tr     = InitTrace(); 
   dolist = InitTracestack();

   delenX = DealignedLength(seqX);
   delenY = DealignedLength(seqY);
   /* attach the guy we'll work on first */
   PushTracestack(dolist, 
	 AttachATrace(tr, 0, 0, len-1, 0, delenX-1, 0, delenY-1, NTRANS));

   switch(grammar) {
      case NUS:
	 khs2traceNUS(dolist, ct, seqX, seqY);
	 break;
      case YRN:
      case YR2:
	 khs2traceYRN(dolist, ct, grammar, seqX, seqY);
	 break;
      case MNG:
	 khs2traceMNG(dolist, ct, seqX, seqY);
	 break;
      case STA:
	 khs2traceSTA(dolist, ct, seqX, seqY);
	 break;
      default: 
	 break;
   }

   if (debug) printTraceGC(stdout, tr);

   *ret_tr = tr;
   FreeTracestack(dolist);
   free(ct);

   if (!validstruct) return 0;
   else return 1;
}

/* Function: KHS2ct()
 * 
 * Purpose:  Convert a secondary structure string to an array of integers
 *           representing what position each position is base-paired 
 *           to (0..len-1), or -1 if none. This is off-by-one from a
 *           Zuker .ct file representation.
 *           
 *           The .ct representation can accomodate pseudoknots but the 
 *           secondary structure string cannot easily; the string contains
 *           "Aa", "Bb", etc. pairs as a limited representation of
 *           pseudoknots. The string contains "><" for base pairs.
 *           Other symbols are ignored. If allow_pseudoknots is FALSE,
 *           the pseudoknot symbols will be ignored and these positions
 *           will be treated as single stranded.
 *           
 * Return:   ret_ct is allocated here and must be free'd by caller.
 *           Returns 1 on success, 0 if ss is somehow inconsistent.
 */
int 
KHS2ct(char *ss, int len, int allow_pseudoknots, int **ret_ct)
{
  struct intstack_s *dolist[27];
  int *ct;
  int  i;
  int  pos, pair;
  int  status = 1;		/* success or failure return status */

  for (i = 0; i < 27; i++)
    dolist[i] = InitIntStack();

  if ((ct = (int *) malloc (len * sizeof(int))) == NULL)
    Die("malloc failed");
  for (pos = 0; pos < len; pos++)
    ct[pos] = -1;

  for (pos = 0; ss[pos] != '\0'; pos++)
    {
      /* bulletproof against SGI non-ANSI compliant ctype.h */
      if (! sre_isascii(ss[pos])) status = 0;

      else if (ss[pos] == '>')	/* left side of a pair: push onto stack 0 */
	PushIntStack(dolist[0], pos);
      else if (ss[pos] == '<')	/* right side of a pair; resolve pair */
	{
	  if (! PopIntStack(dolist[0], &pair))
	    { status = 0; }
	  else
	    {
	      ct[pos]  = pair;
	      ct[pair] = pos;
	    }
	}
				/* same stuff for pseudoknots */
      else if (allow_pseudoknots && isupper((int) ss[pos]))
	PushIntStack(dolist[ss[pos] - 'A' + 1], pos);
      else if (allow_pseudoknots && islower((int) ss[pos]))
	{
	  if (! PopIntStack(dolist[ss[pos] - 'a' + 1], &pair))
	    { status = 0; }
	  else
	    {
	      ct[pos]  = pair;
	      ct[pair] = pos;
	    }
	}
      else if (allow_pseudoknots && !isgap(ss[pos])) 
	status = 0; /* bad character */
    }

  for (i = 0; i < 27; i++)
    if ( FreeIntStack(dolist[i]) > 0)
      status = 0;

  *ret_ct = ct;
  return status;
}

/* Function: VerifyKHS()
 * 
 * Purpose:  Examine a possibly bad structure string, and print out diagnostics 
 *           about it if wordy is TRUE.
 *
 * Return:   1 if string is OK, 0 if string is bad.
 */
int
VerifyKHS(char *name, char *ss, int wordy)
{
  int symcount[27];             /* 0 is normal pairs. 1-26 for pseudoknots */
  int i;
  int pos;
  int status = 1;

  for (i = 0; i < 27; i++)
    symcount[i] = 0;

  for (pos = 0; ss[pos] != '\0'; pos++)
    {
      if (!sre_isascii(ss[pos]))        /* SGI ctype.h non-ANSI compliant */
        {
          status = 0;
          if (wordy)
            Warn("VerifyKHS: Sequence %s no good; structure has garbage symbol (val %d) at position %d\n",
                 name, (int) ss[pos], pos);
        }
      else if (ss[pos] == '>')
        symcount[0] ++;
      else if (ss[pos] == '<')
        symcount[0] --;
      else if (isupper((int) ss[pos]))
        symcount[ss[pos] - 'A' + 1] ++; /* pseudoknot-left  */
      else if (islower((int) ss[pos]))
        symcount[ss[pos] - 'a' + 1] --; /* pseudoknot-right */
      else if (!isgap(ss[pos]))
        {
          status = 0;
          if (wordy)
            Warn("VerifyKHS: Sequence %s no good, structure has invalid symbol %c at position %d\n",
                 name, ss[pos], pos);
        }
    }

  if (symcount[0] != 0)
    {
      status = 0;
      if (wordy)
        Warn("VerifyKHS: Sequence %s no good, structure has extra paired bases: %d on %s\n",
             name, abs(symcount[0]),
             (symcount[0] > 0) ? "left" : "right");
    }

  for (i = 1; i <= 26; i++)
    if (symcount[i] != 0)
      {
        status = 0;
        if (wordy)
          Warn("VerifyKHS: Sequence %s no good, structure has extra paired bases for pseudoknot %c: %d on %s\n",
               name, (char) (i + 'A' - 1),
               abs(symcount[i]),
               (symcount[i] > 0) ? "left" : "right");
      }
  return status;
}

int 
khsSanity(char *seq, char *ss)
{
   /* Given a sequence string and it's structure,
    * verify that the proposed locations of pairs
    * makes sense.
    */
   return 1;
}
